/*
    This library function takes data from the output of mesh
    generating code and writes it out to a format appropriate
    for reading into SLFFEA.  Currently, it can write out data
    for triangle and tetrahedron elements.

		Updated 1/9/07

    SLFFEA source file
    Version:  1.4
    Copyright (C) 2007  San Le 

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 
*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

extern int numel, numnp, triangle_flag;

int slffeawrite(double *p, int *connect, int npel, int nsd)
{
	int dof, nmat, nmode, plane_stress_flag, npel2;
        int i,j,dum,dum2,dum3,dum4;
        char name[20];
	char buf[ BUFSIZ ];
        FILE *o2;
	char text;

	nmat = 1;
	nmode = 0;
	plane_stress_flag = 0;

	npel2 = 4;
	if( triangle_flag ) npel2 = 3;

        o2 = fopen( "slffea.dat","w" );

        fprintf( o2, "   numel numnp nmat nmode");
        if( triangle_flag ) fprintf( o2, " plane_stress_flag\n ");
        else fprintf( o2, "\n ");
        fprintf( o2, "    %4d %4d %4d %4d", numel, numnp, nmat, nmode);
        if( triangle_flag ) fprintf( o2, " %4d\n ", plane_stress_flag);
        else fprintf( o2, "\n ");

        fprintf( o2, "matl no., E modulus, Poisson Ratio, density \n");
        for( i = 0; i < nmat; ++i )
        {
           fprintf( o2, "%3d  %9.4f %9.4f %9.4f\n ", 0, 30e6, 0.3, 2233.0000);
        }

        fprintf( o2, "el no., connectivity, matl no. \n");
        for( i = 0; i < numel; ++i )
        {
           fprintf( o2, "%4d ",i);
           for( j = 0; j < npel2; ++j )
           {
                fprintf( o2, "%4d ", *(connect + npel*i + j));
           }
           fprintf( o2, " %3d\n",0);
        }

        fprintf( o2, "node no., coordinates \n");
        for( i = 0; i < numnp; ++i )
        {
           fprintf( o2, "%4d",i);
           for( j = 0; j < nsd; ++j )
           {
           	fprintf( o2, " %16.8e", *(p + nsd*i + j));
	   }
           fprintf( o2, "\n");
        }

        dum= 0;
        fprintf( o2, "prescribed displacement x: node  disp value\n");
        fprintf( o2, "%4d\n ",-10);
        fprintf( o2, "prescribed displacement y: node  disp value\n");
        fprintf( o2, "%4d\n ",-10);
        if( !triangle_flag )
	{
		fprintf( o2, "prescribed displacement z: node  disp value\n");
        	fprintf( o2, "%4d\n ",-10);
	}

        fprintf( o2, "node with point load and load vector in x,y\n");
        fprintf( o2, "%4d\n ",-10);
        fprintf( o2, "element and gauss pt. with stress and stress vector in xx,yy,xy\n");
        fprintf( o2, "%4d ",-10);

        return 1;
}

