/*
    This program generates meshes for triangle elements.  It does it by
    sub-division.  It is taken from the Matlab functions trgl3_sqr.m and
    trgl3_disk.m in the package FSELIB for chapter 3 written by C. Pozrikidis
    from his book:

 Pozrikidis, C., Introduction to Finite and Spectral Element Methods using Matlab,
      Chapman & Hall/CRC, Boca Raton, 2005.

    You can get the complete FSELIB package at:

      http://dehesa.freeshell.org/FSELIB/

                Updated 1/17/07 Translated into C by San Le

                Copyright C. Pozrikidis

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define SMALL   0.000001 


int ne, ng;
int numnp, numel, shape_flag, triangle_flag = 1, circle_flag;
int npel = 3, nsd = 2;

int trglreader(int *, char *, int , FILE *, double *, double *);

int slffeawrite(double *, int *, int, int);

int trgl3( double *x, double *y, double *p, int *efl, int *c, int ndiv)
{

	int i,i2,j,k,dum,dum2,dum3,dum4;
	double fdum, fdum1, fdum2, fdum3, fdum4, Iflag;
	double *xn, *yn, xp[40], yp[40], rad;
	int *efln, *gfl;

	int nm;

	xn = (double *)calloc(numel*4*ndiv*3,sizeof(double));
	yn = (double *)calloc(numel*4*ndiv*3,sizeof(double));
	efln = (int *)calloc(numel*4*ndiv*3,sizeof(int));
	gfl = (int *)calloc(numnp*3,sizeof(int));


/*
  refinement loop
*/

	if(ndiv > 0)
	{
	    for( i = 0; i < ndiv; ++i )
	    {
		nm = 0; /* count the new elements arising by each refinement loop
			   four elements will be generated in each pass */

		for( j = 0; j < ne; ++j )   /* loop over current elements */
		{

/* edge mid-nodes will become vertex nodes */

		    x[6*j+3] = 0.5*(x[6*j+0]+x[6*j+1]); y[6*j+3] = 0.5*(y[6*j+0]+y[6*j+1]);
		    x[6*j+4] = 0.5*(x[6*j+1]+x[6*j+2]); y[6*j+4] = 0.5*(y[6*j+1]+y[6*j+2]);
		    x[6*j+5] = 0.5*(x[6*j+2]+x[6*j+0]); y[6*j+5] = 0.5*(y[6*j+2]+y[6*j+0]);

		    efl[6*j+3] = 0; if(efl[6*j+0] == 1 & efl[6*j+1] == 1) efl[6*j+3] = 1;
		    efl[6*j+4] = 0; if(efl[6*j+1] == 1 & efl[6*j+2] == 1) efl[6*j+4] = 1;
		    efl[6*j+5] = 0; if(efl[6*j+2] == 1 & efl[6*j+0] == 1) efl[6*j+5] = 1;

		    /*printf( "aaa %4d %4d %16.8e %16.8e %16.8e\n", i, j, x[6*j+3], x[6*j+4], x[6*j+5]);*/

  /* assign vertex nodes to sub-elements
     these will become the "new" elements */

		    xn[3*nm+0] = x[6*j+0]; yn[3*nm+0] = y[6*j+0]; efln[3*nm+0] = efl[6*j+0]; /*  first sub-element */
		    xn[3*nm+1] = x[6*j+3]; yn[3*nm+1] = y[6*j+3]; efln[3*nm+1] = efl[6*j+3];
		    xn[3*nm+2] = x[6*j+5]; yn[3*nm+2] = y[6*j+5]; efln[3*nm+2] = efl[6*j+5];
		    nm += 1;

		    xn[3*nm+0] = x[6*j+3]; yn[3*nm+0] = y[6*j+3]; efln[3*nm+0] = efl[6*j+3]; /*  second sub-element */
		    xn[3*nm+1] = x[6*j+1]; yn[3*nm+1] = y[6*j+1]; efln[3*nm+1] = efl[6*j+1];
		    xn[3*nm+2] = x[6*j+4]; yn[3*nm+2] = y[6*j+4]; efln[3*nm+2] = efl[6*j+4];
		    nm += 1;

		    xn[3*nm+0] = x[6*j+5]; yn[3*nm+0] = y[6*j+5]; efln[3*nm+0] = efl[6*j+5]; /*  third sub-element */
		    xn[3*nm+1] = x[6*j+4]; yn[3*nm+1] = y[6*j+4]; efln[3*nm+1] = efl[6*j+4];
		    xn[3*nm+2] = x[6*j+2]; yn[3*nm+2] = y[6*j+2]; efln[3*nm+2] = efl[6*j+2];
		    nm += 1;

		    xn[3*nm+0] = x[6*j+3]; yn[3*nm+0] = y[6*j+3]; efln[3*nm+0] = efl[6*j+3]; /*  fourth sub-element */
		    xn[3*nm+1] = x[6*j+4]; yn[3*nm+1] = y[6*j+4]; efln[3*nm+1] = efl[6*j+4];
		    xn[3*nm+2] = x[6*j+5]; yn[3*nm+2] = y[6*j+5]; efln[3*nm+2] = efl[6*j+5];
		    nm += 1;


		    /*end of loop over current elements*/
		}


		ne = 4*ne;  /* number of elements has increased by a factor of four */
			    /* If shape is a disk, project boundary nodes onto the unit circle */

		for( k = 0; k < ne; ++k ) /* relabel the new points and put them in the master list */
		{
		    for( i2 = 0; i2 < 3; ++i2 )
		    {
			x[6*k+i2] = xn[3*k+i2];
			y[6*k+i2] = yn[3*k+i2];
			efl[6*k+i2] = efln[3*k+i2];
			if( shape_flag == 3 || circle_flag == 1)
			{
			    if( efl[6*k+i2] == 1 )
			    {
				rad = sqrt(x[6*k+i2]*x[6*k+i2] + y[6*k+i2]*y[6*k+i2]);
				x[6*k+i2] /=rad;
				y[6*k+i2] /=rad;
			    }
			}
		    }
		}
	    } /* end for loop of refinement */
	} /* end if of refinement loop */

	/*for( k = 0; k < ne; ++k ) 
	{
		for( i2 = 0; i2 < 3; ++i2 )
		{
			printf( "mmm %4d %4d %4d %4d %16.8e %16.8e %4d\n",
				3, 128, k+1, i2+1, x[6*k+i2], y[6*k+i2], efl[6*k+i2]);
		}
	}*/

	for( i = 0; i < ne; ++i ) /* relabel the new points
	{
	    xp[0] = x[6*i]; xp[1] = x[6*i+1]; xp[2] = x[6*i+2]; xp[3] = x[6*i];
	    yp[0] = y[6*i]; yp[1] = y[6*i+1]; yp[2] = y[6*i+2]; yp[3] = y[6*i];
	}

/* define the global nodes
   and the connectivity table
*/

/* three nodes of the first element are entered mannualy */

	p[0] = x[0]; p[1] = y[0];  gfl[0] = efl[0];
	p[2] = x[1]; p[3] = y[1];  gfl[3] = efl[1];
	p[4] = x[2]; p[5] = y[2];  gfl[6] = efl[2];

	c[0] = 0;  /* first  node of first element is global node 0 */
	c[1] = 1;  /* second node of first element is global node 1 */
	c[2] = 2;  /* third  node of first element is global node 2 */

/* loop over further elements
   Iflag=0 will signal a new global node
*/

	for( i = 1; i < ne; ++i )  /* loop over elements */
	{
	    for( j = 0; j < 3; ++j )  /* loop over element nodes */
	    {
		Iflag = 0;
		for( k = 0; k < ng; ++k )
		{
		    /*printf( "vvv %4d %4d %4d %16.8e %16.8e %16.8e %16.8e\n",
			i, j, ng, x[6*i+j], p[2*k], y[6*i+j], p[2*k+1]);*/
		   if( fabs( x[6*i+j] - p[2*k] ) < SMALL )
		   {
			if( fabs( y[6*i+j] - p[2*k+1] ) < SMALL )
			{
			    Iflag = 1;    /* the node has been recorded previously */
			    c[npel*i+j] = k;  /* the jth local node of element i is the kth global node */
		    /*printf( "rrr %4d %4d %4d %16.8e %16.8e %16.8e %16.8e %4d\n",
			i, j, ng, x[6*i+j], p[2*k], y[6*i+j], p[2*k+1], c[npel*i+j]);*/
			}
		   }
		}

		if(Iflag == 0)  /* record the node */
		{
		    p[2*ng+0] = x[6*i+j];
		    p[2*ng+1] = y[6*i+j];
		    gfl[3*ng+0] = efl[6*i+j];
		    c[npel*i+j] = ng;   /* the jth local node of element is the new global node */
		    ng += 1;
		}
	    }
	} /* end of loop over elements */

	return 1;
}


/*

1) First problem:

 discretization of a square
 subtended between -1<x<1, -1<y<1
 into 3-node triangles
 
 c: connectivity matrix

 parental structure with eight elements

2) Second problem:

 triangulation of the unit disk
 by successive subdivisions
 of a parental structure

 efl: element flag
 gfl: global node flag

*/

int main(int argc, char** argv)
{

	int i, j, k, check;
	int dum, dum2;
	double *x, *y, *p;
	int *efl, *c, ndiv;
	char name[20];
	char buf[ BUFSIZ ];
	FILE *o1;

	shape_flag = 3;
	printf( "Pick the mesh you want among 3\n");
	printf( "1) Square made up of 2048 triangle elements\n");
	printf( "2) Square made up of 512 triangle elements\n");
	printf( "3) Disk made up of 256 triangle elements\n");
	printf( "4) Use input file\n");
	scanf( "%d", &shape_flag);

	if(shape_flag == 4)
	{
	    ng = 3;
	    printf( "What is the name of the file containing the \n");
	    printf( "initial triangle data? (example: dat3b-3)\n");
	    scanf( "%30s",name);

	    o1 = fopen( name,"r" );
	    if(o1 == NULL ) {
		printf("Can't find file %30s\n",name);
		exit(1);
	    }
	    printf( "   ne ndiv circle_flag\n ");
	    fgets( buf, BUFSIZ, o1 );
	    fscanf( o1, "%d %d %d\n",&ne,&ndiv,&circle_flag);
	    printf( "%4d %4d %4d\n", ne, ndiv, circle_flag);
	}

	if(shape_flag == 3)
	{
/* In scd3_d.m, for ndiv = 3 ne = 4 and ng = 3, numel = 256, numnp = 145 */
	    ne = 4;
	    ng = 3;
	    ndiv = 3;
	}

	if(shape_flag == 2)
	{
/* In lapl3_dn_sqr.m, for ndiv = 3 ne = 8 and ng = 3, numel = 512, numnp = 289 */
	    ne = 8;
	    ng = 3;
	    ndiv = 3;
	}

	if(shape_flag == 1)
	{
/* In pois3_d.m, for ndiv = 4 ne = 8 and ng = 3, numel = 2048, numnp = 1089 */
	    ne = 8;
	    ng = 3;
	    ndiv = 4;
	}

	numel = ne*pow(4,ndiv);
	numnp = npel*numel; /* This is temporary so I can set memory.*/
	printf( " %4d %4d\n", numel, numnp);

	x = (double *)calloc(numel*6,sizeof(double));
	y = (double *)calloc(numel*6,sizeof(double));
	efl = (int *)calloc(numel*6,sizeof(int));
	c = (int *)calloc(numel*npel,sizeof(int));
	p = (double *)calloc(numnp*nsd,sizeof(double));

	if(shape_flag == 4)
	{
	    check = trglreader( efl, name, ndiv, o1, x, y);
	    if(!check) printf( " Problems with trglreader\n");
	}

	if(shape_flag == 3)
	{
	    x[0] = 0.0; y[0] = 0.0; efl[0] = 0;  /* first element */
	    x[1] = 1.0; y[1] = 0.0; efl[1] = 1;
	    x[2] = 0.0; y[2] = 1.0; efl[2] = 1;

	    x[6] = 0.0; y[6] = 0.0; efl[6] = 0;  /* second element */
	    x[7] = 0.0; y[7] = 1.0; efl[7] = 1;
	    x[8] =-1.0; y[8] = 0.0; efl[8] = 1;

	    x[12] = 0.0; y[12] = 0.0; efl[12] = 0;  /* third element */
	    x[13] =-1.0; y[13] = 0.0; efl[13] = 1;
	    x[14] = 0.0; y[14] =-1.0; efl[14] = 1;

	    x[18] = 0.0; y[18] = 0.0; efl[18] = 0;  /* fourth element */
	    x[19] = 0.0; y[19] =-1.0; efl[19] = 1;
	    x[20] = 1.0; y[20] = 0.0; efl[20] = 1;
	}

	if(shape_flag < 3)
	{
	    x[0] = 0.0; y[0] = 0.0; efl[0] = 0;  /* first element */
	    x[1] = 1.0; y[1] = 0.0; efl[1] = 1;
	    x[2] = 1.0; y[2] = 1.0; efl[2] = 1;

	    x[6] = 0.0; y[6] = 0.0; efl[6] = 0;  /* second element */
	    x[7] = 1.0; y[7] = 1.0; efl[7] = 1;
	    x[8] = 0.0; y[8] = 1.0; efl[8] = 1;

	    x[12] = 0.0; y[12] = 0.0; efl[12] = 0;  /* third element */
	    x[13] = 0.0; y[13] = 1.0; efl[13] = 1;
	    x[14] =-1.0; y[14] = 1.0; efl[14] = 1;

	    x[18] = 0.0; y[18] = 0.0; efl[18] = 0;  /* fourth element */
	    x[19] =-1.0; y[19] = 1.0; efl[19] = 1;
	    x[20] =-1.0; y[20] = 0.0; efl[20] = 1;

	    x[24] = 0.0; y[24] = 0.0; efl[24] = 0;  /* fifth element */
	    x[25] =-1.0; y[25] = 0.0; efl[25] = 1;
	    x[26] =-1.0; y[26] =-1.0; efl[26] = 1;

	    x[30] = 0.0; y[30] = 0.0; efl[30] = 0;  /* sixth element */
	    x[31] =-1.0; y[31] =-1.0; efl[31] = 1;
	    x[32] = 0.0; y[32] =-1.0; efl[32] = 1;

	    x[36] = 0.0; y[36] = 0.0; efl[36] = 0;  /* seventh element */
	    x[37] = 0.0; y[37] =-1.0; efl[37] = 1;
	    x[38] = 1.0; y[38] =-1.0; efl[38] = 1;

	    x[42] = 0.0; y[42] = 0.0; efl[42] = 0; /* eighth element */
	    x[43] = 1.0; y[43] =-1.0; efl[43] = 1;
	    x[44] = 1.0; y[44] = 0.0; efl[44] = 1;
	}

	check = trgl3( x, y, p, efl, c, ndiv);
	if(!check) printf( " Problems with trgl3\n");

	numnp = ng;
	numel = ne;

	check = slffeawrite( p, c, npel, nsd);
	if(!check) printf( " Problems with slffeawrite \n");

	return 1;

}

